"""YOLO dataset loader implementation."""

from pathlib import Path
from typing import Optional

from labelformat.formats import YOLOv8ObjectDetectionInput


class YOLODatasetLoader:
    """Loader for YOLO format datasets."""

    def __init__(self, data_yaml_path: str, input_split: str):
        """Initialize the YOLO dataset loader.

        Args:
            data_yaml_path: Path to the YOLO data.yaml file
            input_split: The split to load (train, val, test)
        """
        self.data_yaml_path = Path(data_yaml_path)
        self.input_split = input_split
        self._label_input: Optional[YOLOv8ObjectDetectionInput] = None

    def load(self) -> YOLOv8ObjectDetectionInput:
        """Load the YOLO dataset.

        Returns:
            YOLOv8ObjectDetectionInput: The loaded dataset
        """
        if not self._label_input:
            self._label_input = YOLOv8ObjectDetectionInput(
                input_file=self.data_yaml_path,
                input_split=self.input_split,
            )
        return self._label_input



"""This module contains the FastAPI app configuration."""

from contextlib import asynccontextmanager

from fastapi import Depends, FastAPI, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from sqlalchemy.exc import DataError, IntegrityError, OperationalError
from sqlmodel import Session
from typing_extensions import Annotated

from lightly_purple.dataset.env import PURPLE_DEBUG
from lightly_purple.server.cache import StaticFilesCache
from lightly_purple.server.db import db_manager
from lightly_purple.server.routes import healthz, webapp
from lightly_purple.server.routes.api import (
    annotation,
    annotation_label,
    dataset,
    dataset_tag,
    sample,
)
from lightly_purple.server.routes.api.status import (
    HTTP_STATUS_BAD_REQUEST,
    HTTP_STATUS_CONFLICT,
    HTTP_STATUS_INTERNAL_SERVER_ERROR,
)

SessionDep = Annotated[Session, Depends(db_manager.session)]


@asynccontextmanager
async def lifespan(_: FastAPI):
    """Lifespan context for initializing and cleaning up resources."""
    yield


if PURPLE_DEBUG == "True":
    import logging

    logging.basicConfig()
    logger = logging.getLogger("sqlalchemy.engine")
    logger.setLevel(logging.DEBUG)

"""Create the FastAPI app."""
app = FastAPI(lifespan=lifespan)

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

app.mount(
    "/images",
    StaticFilesCache(directory="/"),
    "images",
)


@app.exception_handler(IntegrityError)
async def _integrity_error_handler(_request, _exc):
    return HTTPException(
        status_code=HTTP_STATUS_CONFLICT, detail="Database constraint violated."
    )


@app.exception_handler(DataError)
async def _data_error_handler(_request, _exc):
    return HTTPException(
        status_code=HTTP_STATUS_BAD_REQUEST, detail="Invalid data provided."
    )


@app.exception_handler(OperationalError)
async def _operational_error_handler(_request, _exc):
    return HTTPException(
        status_code=HTTP_STATUS_INTERNAL_SERVER_ERROR,
        detail="Database operation failed.",
    )


# api routes
app.include_router(dataset.dataset_router, prefix="/api")
app.include_router(dataset_tag.tag_router, prefix="/api")
app.include_router(sample.samples_router, prefix="/api")
app.include_router(annotation_label.annotations_label_router, prefix="/api")
app.include_router(annotation.annotations_router, prefix="/api")

# webapp routes
app.include_router(webapp.app_router)

# health status check
app.include_router(healthz.health_router)



"""This module contains the API routes for managing annotation labels."""

from typing import List
from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException
from sqlmodel import Session
from typing_extensions import Annotated

from lightly_purple.server.db import get_session
from lightly_purple.server.models import AnnotationLabel, AnnotationLabelView
from lightly_purple.server.models.annotation_label import (
    AnnotationLabelInput,
)
from lightly_purple.server.resolvers.annotation_label import (
    AnnotationLabelResolver,
)
from lightly_purple.server.routes.api.status import (
    HTTP_STATUS_CREATED,
    HTTP_STATUS_NOT_FOUND,
)

annotations_label_router = APIRouter()
SessionDep = Annotated[Session, Depends(get_session)]


def get_annotation_label_resolver(
    session: SessionDep,
) -> AnnotationLabelResolver:
    """Create an instance of the AnnotationLabelResolver."""
    return AnnotationLabelResolver(session)


@annotations_label_router.post(
    "/annotation_labels",
    response_model=AnnotationLabelView,
    status_code=HTTP_STATUS_CREATED,
)
def create_annotation_label(
    input_label: AnnotationLabelInput,
    handler: Annotated[
        AnnotationLabelResolver, Depends(get_annotation_label_resolver)
    ],
) -> AnnotationLabel:
    """Create a new annotation label in the database."""
    return handler.create(input_label)


@annotations_label_router.get("/annotation_labels")
def read_annotation_labels(
    handler: Annotated[
        AnnotationLabelResolver, Depends(get_annotation_label_resolver)
    ],
) -> List[AnnotationLabel]:
    """Retrieve a list of annotation labels from the database."""
    return handler.get_all()


@annotations_label_router.get("/annotation_labels/{label_id}")
def read_annotation_label(
    label_id: UUID,
    handler: Annotated[
        AnnotationLabelResolver, Depends(get_annotation_label_resolver)
    ],
) -> AnnotationLabel:
    """Retrieve a single annotation label from the database."""
    label = handler.get_by_id(label_id)
    if not label:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND,
            detail="Annotation label not found",
        )
    return label


@annotations_label_router.put("/annotation_labels/{label_id}")
def update_annotation_label(
    label_id: UUID,
    label_input: AnnotationLabel,
    handler: Annotated[
        AnnotationLabelResolver, Depends(get_annotation_label_resolver)
    ],
) -> AnnotationLabel:
    """Update an existing annotation label in the database."""
    label = handler.update(label_id, label_input)
    if not label:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND,
            detail="Annotation label not found",
        )
    return label


@annotations_label_router.delete("/annotation_labels/{label_id}")
def delete_annotation_label(
    label_id: UUID,
    handler: Annotated[
        AnnotationLabelResolver, Depends(get_annotation_label_resolver)
    ],
):
    """Delete an annotation label from the database."""
    if not handler.delete(label_id):
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND,
            detail="Annotation label not found",
        )
    return {"status": "deleted"}



"""Handler for database operations related to datasets."""

from datetime import datetime, timezone
from typing import List, Optional
from uuid import UUID

from sqlmodel import Session, select

from lightly_purple.server.models import Dataset
from lightly_purple.server.models.dataset import DatasetInput


class DatasetResolver:
    """Resolver for the Dataset model."""

    def __init__(self, session: Session):  # noqa: D107
        self.session = session

    def create(self, dataset: DatasetInput) -> Dataset:
        """Create a new dataset in the database."""
        db_dataset = Dataset.model_validate(dataset)
        self.session.add(db_dataset)
        self.session.commit()
        self.session.refresh(db_dataset)
        return db_dataset

    def get_all(self, offset: int = 0, limit: int = 100) -> List[Dataset]:
        """Retrieve all datasets with pagination."""
        datasets = self.session.exec(
            select(Dataset).offset(offset).limit(limit)
        ).all()
        return list(datasets) if datasets else []

    def get_by_id(self, dataset_id: UUID) -> Optional[Dataset]:
        """Retrieve a single dataset by ID."""
        return self.session.exec(
            select(Dataset).where(Dataset.dataset_id == dataset_id)
        ).one_or_none()

    def update(
        self, dataset_id: UUID, dataset_data: DatasetInput
    ) -> Optional[Dataset]:
        """Update an existing dataset."""
        dataset = self.get_by_id(dataset_id)
        if not dataset:
            return None

        dataset.name = dataset_data.name
        dataset.directory = dataset_data.directory
        dataset.updated_at = datetime.now(timezone.utc)

        self.session.commit()
        self.session.refresh(dataset)
        return dataset

    def delete(self, dataset_id: UUID) -> bool:
        """Delete a dataset."""
        dataset = self.get_by_id(dataset_id)
        if not dataset:
            return False

        self.session.delete(dataset)
        self.session.commit()
        return True



"""General purple API models."""

from pydantic import BaseModel
from sqlmodel import Field


class Paginated(BaseModel):
    """Paginated query parameters."""

    offset: int = Field(0, ge=0, description="Offset for pagination")
    limit: int = Field(100, gt=0, le=100, description="Limit for pagination")



"""Handler for database operations related to annotation labels."""

from typing import List, Optional
from uuid import UUID

from sqlmodel import Session, select

from lightly_purple.server.models import AnnotationLabel
from lightly_purple.server.models.annotation_label import (
    AnnotationLabelInput,
)


class AnnotationLabelResolver:
    """Resolver for the AnnotationLabel model."""

    def __init__(self, session: Session):  # noqa: D107
        self.session = session

    def create(self, label: AnnotationLabelInput) -> AnnotationLabel:
        """Create a new annotation label in the database."""
        db_label = AnnotationLabel.model_validate(label)
        self.session.add(db_label)
        self.session.commit()
        self.session.refresh(db_label)
        return db_label

    def get_all(self) -> List[AnnotationLabel]:
        """Retrieve all annotation labels."""
        labels = self.session.exec(select(AnnotationLabel)).all()
        return list(labels) if labels else []

    def get_by_id(self, label_id: UUID) -> Optional[AnnotationLabel]:
        """Retrieve a single annotation label by ID."""
        return self.session.exec(
            select(AnnotationLabel).where(
                AnnotationLabel.annotation_label_id == label_id
            )
        ).one_or_none()

    def update(
        self, label_id: UUID, label_data: AnnotationLabelInput
    ) -> Optional[AnnotationLabel]:
        """Update an existing annotation label."""
        label = self.get_by_id(label_id)
        if not label:
            return None

        label.annotation_label_name = label_data.annotation_label_name
        self.session.commit()
        self.session.refresh(label)
        return label

    def delete(self, label_id: UUID) -> bool:
        """Delete an annotation label."""
        label = self.get_by_id(label_id)
        if not label:
            return False

        self.session.delete(label)
        self.session.commit()
        return True



"""This module contains the API routes for managing datasets."""

from pathlib import Path

from fastapi import APIRouter, HTTPException
from fastapi.responses import FileResponse

from .api.status import HTTP_STATUS_NOT_FOUND

app_router = APIRouter()

# Get the current project root directory
project_root = Path(__file__).resolve().parent.parent.parent

webapp_dir = project_root / "dist-purple-view-app"

# Check if the directory exists and raise an error if it doesn't
if not webapp_dir.exists():
    raise RuntimeError(
        f"Directory '{webapp_dir}' does not exist in '{project_root}'"
    )

# Ensure the path is absolute
webapp_dir = webapp_dir.resolve()


@app_router.get("/", include_in_schema=False)
async def serve_index_file():
    """Serve the index.html file for the webapp."""
    index_file = webapp_dir / "index.html"
    if not index_file.exists():
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND, detail="Index file not found"
        )
    return FileResponse(index_file)


@app_router.get("/{path:path}", include_in_schema=False)
async def serve_static_files(path: str):
    """Serve static files from the webapp directory."""
    file_path = webapp_dir / path

    if not file_path.exists() or not file_path.is_file():
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND, detail="File not found"
        )
    return FileResponse(file_path)



"""This module defines the User model for the application."""

from typing import Optional
from uuid import UUID

from sqlmodel import Field, SQLModel

from lightly_purple.server.models.sample import SampleViewForAnnotation


class AnnotationBase(SQLModel):
    """Base class for the Annotation model."""

    x: float
    y: float
    width: float
    height: float
    annotation_label_id: UUID = Field(
        default=None,
        index=True,
        foreign_key="annotation_labels.annotation_label_id",
    )
    dataset_id: UUID = Field(
        default=None,
        index=True,
        foreign_key="datasets.dataset_id",
    )
    sample_id: Optional[UUID] = Field(
        default=None,
        foreign_key="samples.sample_id",
    )


class AnnotationInput(AnnotationBase):
    """Annotation class when inserting."""


class AnnotationViewForAnnotationLabel(SQLModel):
    """Annotation view model for annotation."""

    x: float
    y: float
    width: float
    height: float
    sample: "SampleViewForAnnotation" = None



"""This module contains the Tag model and related enumerations."""

from typing import Optional
from uuid import UUID

from sqlmodel import SQLModel


class TagBase(SQLModel):
    """Base class for the Tag model."""

    dataset_id: UUID
    name: str
    description: Optional[str] = None


class TagInput(TagBase):
    """Tag model when creating."""


class TagInputBody(TagInput):
    """Tag model when creating."""

    dataset_id: Optional[UUID] = None  # type: ignore


class TagUpdate(TagBase):
    """Tag model when updating."""


class TagUpdateBody(TagBase):
    """Tag model when updating."""

    dataset_id: Optional[UUID] = None  # type: ignore


class TagView(TagBase):
    """Tag model when retrieving."""

    tag_id: UUID



"""This module contains the Dataset model and related enumerations."""

from uuid import UUID

from sqlmodel import SQLModel


class DatasetBase(SQLModel):
    """Base class for the Dataset model."""

    name: str
    directory: str


class DatasetInput(DatasetBase):
    """Dataset class when inserting."""


class DatasetView(DatasetBase):
    """Dataset class when retrieving."""

    dataset_id: UUID



"""Initialize environment variables for the dataset module."""

import os

from dotenv import load_dotenv

load_dotenv()

PURPLE_PROTOCOL = os.getenv("PURPLE_PROTOCOL")
PURPLE_PORT = os.getenv("PURPLE_PORT")
PURPLE_HOST = os.getenv("PURPLE_HOST")
PURPLE_DEBUG = os.getenv("PURPLE_DEBUG")

APP_URL = f"{PURPLE_PROTOCOL}://{PURPLE_HOST}:{PURPLE_PORT}"



"""Handler for database operations related to tags."""

from datetime import datetime, timezone
from typing import List, Optional
from uuid import UUID

from sqlmodel import Session, delete, select

from lightly_purple.server.models import (
    Annotation,
    AnnotationTagLink,
    Sample,
    SampleTagLink,
    Tag,
)
from lightly_purple.server.models.tag import TagInput, TagUpdate, TagView


class TagResolver:
    """Resolver for the tag model."""

    def __init__(self, session: Session):  # noqa: D107
        self.session = session

    def create(self, tag: TagInput) -> Tag:
        """Create a new tag in the database."""
        db_tag = Tag.model_validate(tag)
        self.session.add(db_tag)
        self.session.commit()
        self.session.refresh(db_tag)
        return db_tag

    def get_all_by_dataset_id(
        self, dataset_id: UUID, offset: int = 0, limit: int = 100
    ) -> List[TagView]:
        """Retrieve all tags with pagination."""
        tags = self.session.exec(
            select(Tag)
            .where(Tag.dataset_id == dataset_id)
            .offset(offset)
            .limit(limit)
        ).all()
        return list(tags) if tags else []

    def get_by_id(self, tag_id: UUID) -> Optional[Tag]:
        """Retrieve a single tag by ID."""
        return self.session.exec(
            select(Tag).where(Tag.tag_id == tag_id)
        ).one_or_none()

    def get_by_name(self, tag_name: str) -> Optional[Tag]:
        """Retrieve a single tag by ID."""
        return self.session.exec(
            select(Tag).where(Tag.tag_name == tag_name)
        ).one_or_none()

    def update(self, tag_id: UUID, tag_data: TagUpdate) -> Optional[Tag]:
        """Update an existing tag."""
        tag = self.get_by_id(tag_id)
        if not tag:
            return None

        # due to duckdb/OLAP optimisations, update operations effecting unique
        # constraints (e.g colums) will lead to a unique constraint violation.
        # This is due to a update is implemented as delete+insert. The error
        # happens only within the same session.
        # To fix it, we can delete, commit + insert a new tag.
        # https://duckdb.org/docs/sql/indexes#over-eager-unique-constraint-checking
        self.session.delete(tag)
        self.session.commit()

        # create clone of tag with updated values
        tag_updated = Tag.model_validate(tag)
        tag_updated.dataset_id = tag_data.dataset_id
        tag_updated.name = tag_data.name
        tag_updated.description = tag_data.description
        tag_updated.updated_at = datetime.now(timezone.utc)

        self.session.add(tag_updated)
        self.session.commit()
        self.session.refresh(tag_updated)
        return tag_updated

    def delete(self, tag_id: UUID) -> bool:
        """Delete a tag."""
        tag = self.get_by_id(tag_id)
        if not tag:
            return False

        self.session.delete(tag)
        self.session.commit()
        return True

    def add_tag_to_sample(
        self,
        tag_id: UUID,
        sample: Sample,
    ) -> Optional[Sample]:
        """Add a tag to a sample."""
        tag = self.get_by_id(tag_id)
        if not tag or not tag.tag_id:
            return None

        sample.tags.append(tag)
        self.session.add(sample)
        self.session.commit()
        self.session.refresh(sample)
        return sample

    def remove_tag_from_sample(
        self,
        tag_id: UUID,
        sample: Sample,
    ) -> Optional[Sample]:
        """Remove a tag from a sample."""
        tag = self.get_by_id(tag_id)
        if not tag or not tag.tag_id:
            return None

        sample.tags.remove(tag)
        self.session.add(sample)
        self.session.commit()
        self.session.refresh(sample)
        return sample

    def add_tag_to_annotation(
        self,
        tag_id: UUID,
        annotation: Annotation,
    ) -> Optional[Annotation]:
        """Add a tag to a annotation."""
        tag = self.get_by_id(tag_id)
        if not tag or not tag.tag_id:
            return None

        annotation.tags.append(tag)
        self.session.add(annotation)
        self.session.commit()
        self.session.refresh(annotation)
        return annotation

    def remove_tag_from_annotation(
        self,
        tag_id: UUID,
        annotation: Annotation,
    ) -> Optional[Annotation]:
        """Remove a tag from a annotation."""
        tag = self.get_by_id(tag_id)
        if not tag or not tag.tag_id:
            return None

        annotation.tags.remove(tag)
        self.session.add(annotation)
        self.session.commit()
        self.session.refresh(annotation)
        return annotation

    def add_sample_ids_to_tag_id(
        self,
        tag_id: UUID,
        sample_ids: List[UUID],
    ) -> Optional[Tag]:
        """Add a list of sample_ids to a tag."""
        tag = self.get_by_id(tag_id)
        if not tag or not tag.tag_id:
            return None

        for sample_id in sample_ids:
            self.session.add(SampleTagLink(sample_id=sample_id, tag_id=tag_id))

        self.session.commit()
        self.session.refresh(tag)
        return tag

    def remove_sample_ids_from_tag_id(
        self,
        tag_id: UUID,
        sample_ids: List[UUID],
    ) -> Optional[Tag]:
        """Remove a list of sample_ids to a tag."""
        tag = self.get_by_id(tag_id)
        if not tag or not tag.tag_id:
            return None

        self.session.exec(
            delete(SampleTagLink).where(
                SampleTagLink.tag_id == tag_id,
                SampleTagLink.sample_id.in_(sample_ids),
            )
        )

        self.session.commit()
        self.session.refresh(tag)
        return tag

    def add_annotation_ids_to_tag_id(
        self,
        tag_id: UUID,
        annotation_ids: List[UUID],
    ) -> Optional[Tag]:
        """Add a list of annotation_ids to a tag."""
        tag = self.get_by_id(tag_id)
        if not tag or not tag.tag_id:
            return None

        for annotation_id in annotation_ids:
            self.session.add(
                AnnotationTagLink(
                    tag_id=tag_id,
                    annotation_id=annotation_id,
                )
            )

        self.session.commit()
        self.session.refresh(tag)
        return tag

    def remove_annotation_ids_from_tag_id(
        self,
        tag_id: UUID,
        annotation_ids: List[UUID],
    ) -> Optional[Tag]:
        """Remove a list of things to a tag."""
        tag = self.get_by_id(tag_id)
        if not tag or not tag.tag_id:
            return None

        self.session.exec(
            delete(AnnotationTagLink).where(
                AnnotationTagLink.tag_id == tag_id,
                AnnotationTagLink.annotation_id.in_(annotation_ids),
            )
        )

        self.session.commit()
        self.session.refresh(tag)
        return tag



"""This module contains the API routes for managing annotations."""

from typing import Dict, List, Optional, Union
from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException
from fastapi.params import Query
from sqlmodel import Session
from typing_extensions import Annotated

from lightly_purple.server.db import get_session
from lightly_purple.server.models.annotation import (
    AnnotationViewForAnnotationLabel,
)
from lightly_purple.server.resolvers.annotation import AnnotationResolver
from lightly_purple.server.resolvers.tag import TagResolver
from lightly_purple.server.routes.api.dataset_tag import get_tag_resolver
from lightly_purple.server.routes.api.status import (
    HTTP_STATUS_CREATED,
    HTTP_STATUS_NOT_FOUND,
)
from lightly_purple.server.routes.api.validators import Paginated

annotations_router = APIRouter()
SessionDep = Annotated[Session, Depends(get_session)]


def get_annotation_resolver(session: SessionDep) -> AnnotationResolver:
    """Create an instance of the AnnotationResolver."""
    return AnnotationResolver(session)


@annotations_router.get("/annotations/count")
def count_annotations_by_dataset(  # noqa: PLR0913 // FIXME: refactor to use proper pydantic
    dataset_id: UUID,
    handler: Annotated[AnnotationResolver, Depends(get_annotation_resolver)],
    filtered_labels: Annotated[Optional[List[str]], Query()] = None,
    min_width: Annotated[Optional[int], Query(ge=0)] = None,
    max_width: Annotated[Optional[int], Query(ge=0)] = None,
    min_height: Annotated[Optional[int], Query(ge=0)] = None,
    max_height: Annotated[Optional[int], Query(ge=0)] = None,
    tag_ids: Optional[List[UUID]] = None,
) -> List[Dict[str, Union[str, int]]]:
    """Get annotation counts for a specific dataset.

    Returns a list of dictionaries with label name and count.
    """
    counts = handler.count_annotations_by_dataset(
        dataset_id,
        filtered_labels,
        min_width=min_width,
        max_width=max_width,
        min_height=min_height,
        max_height=max_height,
        tag_ids=tag_ids,
    )
    return [
        {
            "label_name": label_name,
            "current_count": current_count,
            "total_count": total_count,
        }
        for label_name, current_count, total_count in counts
    ]


@annotations_router.get("/datasets/{dataset_id}/annotations")
def read_annotations(
    dataset_id: UUID,
    handler: Annotated[AnnotationResolver, Depends(get_annotation_resolver)],
    pagination: Annotated[Paginated, Depends()],
    annotation_label_ids: Annotated[Optional[List[UUID]], Query()] = None,
    tag_ids: Optional[List[UUID]] = None,
) -> List[AnnotationViewForAnnotationLabel]:
    """Retrieve a list of annotations from the database."""
    return handler.get_all(
        pagination={
            "offset": pagination.offset,
            "limit": pagination.limit,
        },
        filters={
            "dataset_ids": [dataset_id],
            "annotation_label_ids": annotation_label_ids or None,
            "tag_ids": tag_ids,
        },
    )


@annotations_router.post(
    "/annotations/{annotation_id}/tag/{tag_id}",
    status_code=HTTP_STATUS_CREATED,
)
def add_tag_to_annotation(
    annotation_handler: Annotated[
        AnnotationResolver, Depends(get_annotation_resolver)
    ],
    tag_handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    annotation_id: UUID,
    tag_id: UUID,
) -> bool:
    """Add annotation to a tag."""
    annotation = annotation_handler.get_by_id(annotation_id)
    if not annotation:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND,
            detail=f"Annotation {annotation_id} not found",
        )

    if not tag_handler.add_tag_to_annotation(tag_id, annotation):
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND, detail=f"Tag {tag_id} not found"
        )

    return True


@annotations_router.delete("/annotations/{annotation_id}/tag/{tag_id}")
def remove_tag_from_annotation(
    annotation_handler: Annotated[
        AnnotationResolver, Depends(get_annotation_resolver)
    ],
    tag_handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    tag_id: UUID,
    annotation_id: UUID,
) -> bool:
    """Remove annotation from a tag."""
    annotation = annotation_handler.get_by_id(annotation_id)
    if not annotation:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND,
            detail=f"Annotation {annotation_id} not found",
        )

    if not tag_handler.remove_tag_from_annotation(tag_id, annotation):
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND, detail=f"Tag {tag_id} not found"
        )

    return True



"""This module defines the AnnotationLabel model for the application."""

from sqlmodel import SQLModel


class AnnotationLabelBase(SQLModel):
    """Base class for the AnnotationLabel model."""

    annotation_label_name: str


class AnnotationLabelInput(AnnotationLabelBase):
    """AnnotationLabel class when inserting."""



"""This module contains the API routes for managing datasets."""

from typing import List
from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException, Query
from sqlmodel import Session
from typing_extensions import Annotated

from lightly_purple.server.db import get_session
from lightly_purple.server.models import Dataset
from lightly_purple.server.models.dataset import DatasetInput, DatasetView
from lightly_purple.server.resolvers.dataset import DatasetResolver
from lightly_purple.server.routes.api.status import HTTP_STATUS_NOT_FOUND
from lightly_purple.server.routes.api.validators import Paginated

dataset_router = APIRouter()
SessionDep = Annotated[Session, Depends(get_session)]


def get_dataset_resolver(session: SessionDep) -> DatasetResolver:
    """Create an instance of the DatasetResolver."""
    return DatasetResolver(session)


@dataset_router.post(
    "/datasets",
    response_model=DatasetView,
    status_code=201,
)
def create_dataset(
    dataset_input: DatasetInput,
    handler: Annotated[DatasetResolver, Depends(get_dataset_resolver)],
) -> Dataset:
    """Create a new dataset in the database."""
    return handler.create(dataset_input)


@dataset_router.get("/datasets")
def read_datasets(
    handler: Annotated[DatasetResolver, Depends(get_dataset_resolver)],
    paginated: Annotated[Paginated, Query()],
) -> List[DatasetView]:
    """Retrieve a list of datasets from the database."""
    return handler.get_all(**paginated.model_dump())


@dataset_router.get("/datasets/{dataset_id}")
def read_dataset(
    dataset_id: UUID,
    handler: Annotated[DatasetResolver, Depends(get_dataset_resolver)],
) -> Dataset:
    """Retrieve a single dataset from the database."""
    dataset = handler.get_by_id(dataset_id)
    if not dataset:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND, detail="Dataset not found"
        )
    return dataset


@dataset_router.put("/datasets/{dataset_id}")
def update_dataset(
    dataset_id: UUID,
    dataset_input: DatasetInput,
    handler: Annotated[DatasetResolver, Depends(get_dataset_resolver)],
) -> Dataset:
    """Update an existing dataset in the database."""
    dataset = handler.update(dataset_id, dataset_input)
    if not dataset:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND, detail="Dataset not found"
        )
    return dataset


@dataset_router.delete("/datasets/{dataset_id}")
def delete_dataset(
    dataset_id: UUID,
    handler: Annotated[DatasetResolver, Depends(get_dataset_resolver)],
):
    """Delete a dataset from the database."""
    if not handler.delete(dataset_id):
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND, detail="Dataset not found"
        )
    return {"status": "deleted"}



"""This module contains the API routes for managing datasets."""

from fastapi import APIRouter

health_router = APIRouter()


@health_router.get("/healthz", include_in_schema=False)
def health_check():
    """Health check endpoint to verify the service is running."""
    return {"status": "healthy"}



"""Handler for database operations related to annotations."""

from typing import List, Optional, Tuple, Union
from uuid import UUID

from pydantic import BaseModel, Field
from sqlmodel import Session, func, select

from lightly_purple.server.models import Annotation, AnnotationLabel, Sample, Tag
from lightly_purple.server.models.annotation import AnnotationInput
from lightly_purple.server.routes.api.validators import Paginated


class AnnotationsFilterParams(BaseModel):
    """Encapsulates filter parameters for querying annotations."""

    dataset_ids: Optional[List[UUID]] = Field(
        default=None, description="List of dataset UUIDs"
    )
    annotation_label_ids: Optional[List[UUID]] = Field(
        default=None, description="List of annotation label UUIDs"
    )
    tag_ids: Optional[List[UUID]] = Field(
        default=None, description="List of tag UUIDs"
    )


class AnnotationResolver:
    """Resolver for the Annotation model."""

    def __init__(self, session: Session):  # noqa: D107
        self.session = session

    def create(self, annotation: AnnotationInput) -> Annotation:
        """Create a new annotation in the database."""
        db_annotation = Annotation.model_validate(annotation)
        self.session.add(db_annotation)
        self.session.commit()
        self.session.refresh(db_annotation)
        return db_annotation

    def create_many(self, annotations: List[AnnotationInput]) -> None:
        """Create many annotations in a single commit. No return values."""
        db_annotations = [Annotation.model_validate(a) for a in annotations]
        self.session.bulk_save_objects(db_annotations)
        self.session.commit()

    def get_all(
        self,
        pagination: Optional[Union[dict, Paginated]] = None,
        filters: Optional[AnnotationsFilterParams] = None,
    ) -> List[Annotation]:
        """Retrieve all annotations from the database."""
        if pagination is None:
            pagination = Paginated()
        elif isinstance(pagination, dict):
            pagination = Paginated(**pagination)

        if filters is None:
            filters = AnnotationsFilterParams()
        elif isinstance(filters, dict):
            filters = AnnotationsFilterParams(**filters)

        query = select(Annotation)

        # Apply filters if provided
        if filters:
            if filters.annotation_label_ids:
                query = query.where(
                    Annotation.annotation_label_id.in_(
                        filters.annotation_label_ids
                    )
                )

            if filters.dataset_ids:
                query = query.where(
                    Annotation.dataset_id.in_(filters.dataset_ids)
                )

            # filter by tag_ids
            if filters.tag_ids:
                query = (
                    query.join(Annotation.tags)
                    .where(Annotation.tags.any(Tag.tag_id.in_(filters.tag_ids)))
                    .distinct()
                )

        annotations = self.session.exec(
            query.offset(pagination.offset).limit(pagination.limit)
        ).all()
        return annotations or []

    def get_by_id(self, annotation_id: UUID) -> Optional[Annotation]:
        """Retrieve a single annotation by ID."""
        return self.session.exec(
            select(Annotation).where(Annotation.annotation_id == annotation_id)
        ).one_or_none()

    def update(
        self, annotation_id: UUID, annotation_data: AnnotationInput
    ) -> Optional[Annotation]:
        """Update an existing annotation."""
        annotation = self.get_by_id(annotation_id)
        if not annotation:
            return None

        annotation.x = annotation_data.x
        annotation.y = annotation_data.y
        annotation.width = annotation_data.width
        annotation.height = annotation_data.height

        self.session.commit()
        self.session.refresh(annotation)
        return annotation

    def delete(self, annotation_id: UUID) -> bool:
        """Delete an annotation."""
        annotation = self.get_by_id(annotation_id)
        if not annotation:
            return False

        self.session.delete(annotation)
        self.session.commit()
        return True

    def count_annotations_by_dataset(  # noqa: PLR0913 // FIXME: refactor to use proper pydantic
        self,
        dataset_id: UUID,
        filtered_labels: Optional[List[str]] = None,
        min_width: Optional[int] = None,
        max_width: Optional[int] = None,
        min_height: Optional[int] = None,
        max_height: Optional[int] = None,
        tag_ids: Optional[List[UUID]] = None,
    ) -> List[Tuple[str, int, int]]:
        """Count annotations for a specific dataset.

        Annotations for a specific dataset are grouped by annotation
        label name and counted for total and filtered.
        """
        # Query for total counts (unfiltered)
        total_counts_query = (
            select(
                AnnotationLabel.annotation_label_name,
                func.count(Annotation.annotation_id).label("total_count"),
            )
            .join(
                Annotation,
                Annotation.annotation_label_id
                == AnnotationLabel.annotation_label_id,
            )
            .join(Sample, Sample.sample_id == Annotation.sample_id)
            .where(Sample.dataset_id == dataset_id)
            .group_by(AnnotationLabel.annotation_label_name)
            .order_by(AnnotationLabel.annotation_label_name.asc())
        )

        total_counts = {
            row[0]: row[1]
            for row in self.session.exec(total_counts_query).all()
        }

        # Build filtered query for current counts
        filtered_query = (
            select(
                AnnotationLabel.annotation_label_name,
                func.count(Annotation.annotation_id).label("current_count"),
            )
            .join(
                Annotation,
                Annotation.annotation_label_id
                == AnnotationLabel.annotation_label_id,
            )
            .join(Sample, Sample.sample_id == Annotation.sample_id)
            .where(Sample.dataset_id == dataset_id)
        )

        # Add dimension filters
        if min_width is not None:
            filtered_query = filtered_query.where(Sample.width >= min_width)
        if max_width is not None:
            filtered_query = filtered_query.where(Sample.width <= max_width)
        if min_height is not None:
            filtered_query = filtered_query.where(Sample.height >= min_height)
        if max_height is not None:
            filtered_query = filtered_query.where(Sample.height <= max_height)

        # Add label filter if specified
        if filtered_labels:
            filtered_query = filtered_query.where(
                Sample.sample_id.in_(
                    select(Sample.sample_id)
                    .join(Annotation, Sample.sample_id == Annotation.sample_id)
                    .join(
                        AnnotationLabel,
                        Annotation.annotation_label_id
                        == AnnotationLabel.annotation_label_id,
                    )
                    .where(
                        AnnotationLabel.annotation_label_name.in_(
                            filtered_labels
                        )
                    )
                )
            )

        # filter by tag_ids
        if tag_ids:
            filtered_query = (
                filtered_query.join(Annotation.tags)
                .where(Annotation.tags.any(Tag.tag_id.in_(tag_ids)))
                .distinct()
            )

        # Group by label name and sort
        filtered_query = filtered_query.group_by(
            AnnotationLabel.annotation_label_name
        ).order_by(AnnotationLabel.annotation_label_name.asc())

        current_counts = {
            row[0]: row[1] for row in self.session.exec(filtered_query).all()
        }

        return [
            (label, current_counts.get(label, 0), total_count)
            for label, total_count in total_counts.items()
        ]



"""Handler for database operations related to samples."""

from datetime import datetime, timezone
from typing import Dict, List, Optional
from uuid import UUID

from sqlmodel import Session, func, select

from lightly_purple.server.models import Annotation, AnnotationLabel, Sample, Tag
from lightly_purple.server.models.sample import SampleInput


class SampleResolver:
    """Resolver for the Sample model."""

    def __init__(self, session: Session):  # noqa: D107
        self.session = session

    def create(self, sample: SampleInput) -> Sample:
        """Create a new sample in the database."""
        db_sample = Sample.model_validate(sample)
        self.session.add(db_sample)
        self.session.commit()
        self.session.refresh(db_sample)
        return db_sample

    def get_by_id(self, sample_id: UUID) -> Optional[Sample]:
        """Retrieve a single sample by ID."""
        return self.session.exec(
            select(Sample).where(Sample.sample_id == sample_id)
        ).one_or_none()

    def get_all_by_dataset_id(  # noqa: PLR0913 // FIXME: refactor to use proper pydantic
        self,
        dataset_id: UUID,
        offset: int = 0,
        limit: int = 10,
        min_width: Optional[int] = None,
        max_width: Optional[int] = None,
        min_height: Optional[int] = None,
        max_height: Optional[int] = None,
        annotation_labels: Optional[List[str]] = None,
        tag_ids: Optional[List[UUID]] = None,
    ) -> List[Sample]:
        """Retrieve samples for a specific dataset with optional filtering."""
        query = select(Sample).where(Sample.dataset_id == dataset_id)
        # Add dimension filters
        if min_width is not None:
            query = query.where(Sample.width >= min_width)
        if max_width is not None:
            query = query.where(Sample.width <= max_width)
        if min_height is not None:
            query = query.where(Sample.height >= min_height)
        if max_height is not None:
            query = query.where(Sample.height <= max_height)

        if annotation_labels:
            query = (
                query.join(Annotation, Sample.sample_id == Annotation.sample_id)
                .join(
                    AnnotationLabel,
                    Annotation.annotation_label_id
                    == AnnotationLabel.annotation_label_id,
                )
                .where(
                    AnnotationLabel.annotation_label_name.in_(annotation_labels)
                )
                .distinct()
            )

        if tag_ids:
            query = (
                query.join(Sample.tags)
                .where(Sample.tags.any(Tag.tag_id.in_(tag_ids)))
                .distinct()
            )

        query = (
            query.order_by(Sample.created_at.asc()).offset(offset).limit(limit)
        )
        return self.session.exec(query).all()

    def get_dimension_bounds(
        self,
        dataset_id: UUID,
        annotation_labels: Optional[List[str]] = None,
        tag_ids: Optional[List[UUID]] = None,
    ) -> Dict[str, int]:
        """Get min and max dimensions of samples in a dataset."""
        # Prepare the base query for dimensions
        query = select(
            func.min(Sample.width).label("min_width"),
            func.max(Sample.width).label("max_width"),
            func.min(Sample.height).label("min_height"),
            func.max(Sample.height).label("max_height"),
        )

        if annotation_labels:
            # Subquery to filter samples matching all annotation labels
            label_filter = (
                select(Sample.sample_id)
                .join(Annotation, Sample.sample_id == Annotation.sample_id)
                .join(
                    AnnotationLabel,
                    Annotation.annotation_label_id
                    == AnnotationLabel.annotation_label_id,
                )
                .where(
                    Sample.dataset_id == dataset_id,
                    AnnotationLabel.annotation_label_name.in_(
                        annotation_labels
                    ),
                )
                .group_by(Sample.sample_id)
                .having(
                    func.count(AnnotationLabel.annotation_label_name.distinct())
                    == len(annotation_labels)
                )
            )
            # Filter the dimension query based on the subquery
            query = query.where(Sample.sample_id.in_(label_filter))
        else:
            # If no labels specified, filter dimensions
            # for all samples in the dataset
            query = query.where(Sample.dataset_id == dataset_id)

        if tag_ids:
            query = (
                query.join(Sample.tags)
                .where(Sample.tags.any(Tag.tag_id.in_(tag_ids)))
                .distinct()
            )

        result = self.session.exec(query).one()
        return {
            key: value
            for key, value in result._asdict().items()
            if value is not None
        }

    def update(
        self, sample_id: UUID, sample_data: SampleInput
    ) -> Optional[Sample]:
        """Update an existing sample."""
        sample = self.get_by_id(sample_id)
        if not sample:
            return None

        sample.file_name = sample_data.file_name
        sample.width = sample_data.width
        sample.height = sample_data.height
        sample.updated_at = datetime.now(timezone.utc)

        self.session.commit()
        self.session.refresh(sample)
        return sample

    def delete(self, sample_id: UUID) -> bool:
        """Delete a sample."""
        sample = self.get_by_id(sample_id)
        if not sample:
            return False

        self.session.delete(sample)
        self.session.commit()
        return True



"""This module contains the Server class for running the API using Uvicorn."""

import uvicorn

from lightly_purple.server.app import app


class Server:
    """This class represents a server for running the API using Uvicorn."""

    port: int
    host: str

    def __init__(self, host: str, port: int):
        """Initialize the Server with host and port.

        Args:
            host (str): The hostname to bind the server to.
            port (int): The port number to run the server on.
        """
        self.host = host
        self.port = port

    def start(self):
        """Start the API server using Uvicorn."""
        # start the app
        uvicorn.run(app, host=self.host, port=self.port, http="h11")



"""This module defines the User model for the application."""

from uuid import UUID

from sqlmodel import Field, SQLModel


class SampleBase(SQLModel):
    """Base class for the Sample model."""

    """The name of the image file."""
    file_name: str

    """The width of the image in pixels."""
    width: int

    """The height of the image in pixels."""
    height: int

    """The dataset ID to which the sample belongs."""
    dataset_id: UUID = Field(default=None, foreign_key="datasets.dataset_id")

    """The dataset image path."""
    file_path_abs: str


class SampleInput(SampleBase):
    """Sample class when inserting."""


class SampleViewForAnnotation(SQLModel):
    """Sample class for annotation view."""

    """The name of the image file."""
    file_path_abs: str
    sample_id: UUID



"""HTTP status codes for use in tests and responses."""

HTTP_STATUS_OK = 200
HTTP_STATUS_CREATED = 201
HTTP_STATUS_ACCEPTED = 202
HTTP_STATUS_NO_CONTENT = 204

HTTP_STATUS_BAD_REQUEST = 400
HTTP_STATUS_UNAUTHORIZED = 401
HTTP_STATUS_FORBIDDEN = 403
HTTP_STATUS_NOT_FOUND = 404
HTTP_STATUS_CONFLICT = 409
HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE = 415
HTTP_STATUS_UNPRECESSABLE_ENTITY = 422

HTTP_STATUS_INTERNAL_SERVER_ERROR = 500
HTTP_STATUS_NOT_IMPLEMENTED = 501
HTTP_STATUS_BAD_GATEWAY = 502
HTTP_STATUS_SERVICE_UNAVAILABLE = 503



"""This module contains the API routes for managing samples."""

from typing import Dict, List, Optional
from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException, Path, Query
from pydantic import BaseModel, Field
from sqlmodel import Session
from typing_extensions import Annotated

from lightly_purple.server.db import get_session
from lightly_purple.server.models import Annotation, Sample, SampleView
from lightly_purple.server.models.sample import (
    SampleInput,
)
from lightly_purple.server.resolvers.annotation import AnnotationResolver
from lightly_purple.server.resolvers.sample import SampleResolver
from lightly_purple.server.resolvers.tag import TagResolver
from lightly_purple.server.routes.api.annotation import get_annotation_resolver
from lightly_purple.server.routes.api.dataset_tag import get_tag_resolver
from lightly_purple.server.routes.api.status import (
    HTTP_STATUS_CREATED,
    HTTP_STATUS_NOT_FOUND,
)
from lightly_purple.server.routes.api.validators import Paginated

samples_router = APIRouter()
SessionDep = Annotated[Session, Depends(get_session)]


def get_sample_resolver(session: SessionDep) -> SampleResolver:
    """Create an instance of the SampleResolver."""
    return SampleResolver(session)


@samples_router.post("/samples", response_model=SampleView)
def create_sample(
    handler: Annotated[SampleResolver, Depends(get_sample_resolver)],
    input_sample: SampleInput,
) -> Sample:
    """Create a new sample in the database."""
    return handler.create(input_sample)


# Define additional query parameters model
class SubqueryFilterParams(BaseModel):
    """Query parameters for filtering and doing slice and dice with samples."""

    # metadata
    min_width: Optional[int] = Field(None, ge=0, description="Minimum width")
    max_width: Optional[int] = Field(None, ge=0, description="Maximum width")
    min_height: Optional[int] = Field(None, ge=0, description="Minimum height")
    max_height: Optional[int] = Field(None, ge=0, description="Maximum height")

    # labels
    annotation_labels: Optional[List[str]] = Field(
        None, description="Annotation labels"
    )

    # tags
    tag_ids: Optional[List[UUID]] = Field(
        None, description="Tag IDs to filter by"
    )


# TODO: this should be prefixed with /datasets/{dataset_id}/samples
class ReadSamplesQuery(SubqueryFilterParams, Paginated):
    """Query parameters for reading samples."""

    dataset_id: UUID = Field(description="Dataset ID")


@samples_router.get("/samples")
def read_samples(
    handler: Annotated[SampleResolver, Depends(get_sample_resolver)],
    query: Annotated[ReadSamplesQuery, Query()],
) -> List[SampleView]:
    """Retrieve a list of samples from the database with optional filtering."""
    return handler.get_all_by_dataset_id(
        **query.model_dump(),
    )


@samples_router.get("/samples/dimensions")
def get_sample_dimensions(
    handler: Annotated[SampleResolver, Depends(get_sample_resolver)],
    dataset_id: UUID,
    annotation_labels: Annotated[Optional[List[str]], Query()] = None,
) -> Dict[str, int]:
    """Get min and max dimensions of samples in a dataset."""
    return handler.get_dimension_bounds(
        dataset_id, annotation_labels=annotation_labels
    )


@samples_router.get("/samples/{sample_id}")
def read_sample(
    handler: Annotated[SampleResolver, Depends(get_sample_resolver)],
    sample_id: Annotated[UUID, Path(title="Sample Id")],
) -> SampleView:
    """Retrieve a single sample from the database."""
    sample = handler.get_by_id(sample_id)
    if not sample:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND, detail="Sample not found"
        )

    return sample


@samples_router.get("/samples/{sample_id}/annotations")
def read_sample_annotations(
    handler: Annotated[AnnotationResolver, Depends(get_annotation_resolver)],
    sample_id: Annotated[UUID, Path(title="Sample Id")],
) -> List[Annotation]:
    """Retrieve annotations for sample."""
    return handler.get_by_sample_id(sample_id)


@samples_router.put("/samples/{sample_id}")
def update_sample(
    handler: Annotated[SampleResolver, Depends(get_sample_resolver)],
    sample_id: Annotated[UUID, Path(title="Sample Id")],
    sample_input: Sample,
) -> Sample:
    """Update an existing sample in the database."""
    sample = handler.update(sample_id, sample_input)
    if not sample:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND, detail="Sample not found"
        )
    return sample


@samples_router.delete("/samples/{sample_id}")
def delete_sample(
    handler: Annotated[SampleResolver, Depends(get_sample_resolver)],
    sample_id: Annotated[UUID, Path(title="Sample Id")],
):
    """Delete a sample from the database."""
    if not handler.delete(sample_id):
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND, detail="Sample not found"
        )
    return {"status": "deleted"}


@samples_router.post(
    "/samples/{sample_id}/tag/{tag_id}",
    status_code=HTTP_STATUS_CREATED,
)
def add_tag_to_sample(
    sample_handler: Annotated[SampleResolver, Depends(get_sample_resolver)],
    tag_handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    sample_id: UUID,
    tag_id: UUID,
) -> bool:
    """Add sample to a tag."""
    sample = sample_handler.get_by_id(sample_id)
    if not sample:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND,
            detail=f"Sample {sample_id} not found",
        )

    if not tag_handler.add_tag_to_sample(tag_id, sample):
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND, detail=f"Tag {tag_id} not found"
        )

    return True


@samples_router.delete("/samples/{sample_id}/tag/{tag_id}")
def remove_tag_from_sample(
    sample_handler: Annotated[SampleResolver, Depends(get_sample_resolver)],
    tag_handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    tag_id: UUID,
    sample_id: UUID,
) -> bool:
    """Remove sample from a tag."""
    sample = sample_handler.get_by_id(sample_id)
    if not sample:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND,
            detail=f"Sample {sample_id} not found",
        )

    if not tag_handler.remove_tag_from_sample(tag_id, sample):
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND, detail=f"Tag {tag_id} not found"
        )

    return True



"""This module contains the FastAPI cache configuration for static files."""

from datetime import datetime, timedelta, timezone

from fastapi import Response
from fastapi.staticfiles import StaticFiles

from .routes.api.status import HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE


class StaticFilesCache(StaticFiles):
    """StaticFiles class with cache headers."""

    days_to_expire = 365

    def __init__(
        self,
        *args,
        cachecontrol=f"private, max-age={days_to_expire * 24 * 60 * 60}",
        **kwargs,
    ):
        """Initialize the StaticFilesCache class."""
        self.cachecontrol = cachecontrol
        super().__init__(*args, **kwargs)

    def file_response(self, *args, **kwargs) -> Response:
        """Override the file_response method to add cache headers."""
        allowed_extensions = (
            # Images
            ".png",
            ".jpg",
            ".jpeg",
            ".gif",
            ".webp",
            ".bmp",
            ".gif",
            ".tiff",
            # Movies
            ".mov",
            ".mp4",
            ".avi",
        )

        if not args[0].lower().endswith(allowed_extensions):
            return Response(
                status_code=HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE
            )  # Unsupported Media Type
        resp: Response = super().file_response(*args, **kwargs)
        resp.headers.setdefault("Cache-Control", self.cachecontrol)

        # Calculate expiration date
        expire_date = datetime.now(timezone.utc) + timedelta(
            days=self.days_to_expire
        )
        resp.headers.setdefault(
            "Expires", expire_date.strftime("%a, %d %b %Y %H:%M:%S GMT")
        )

        return resp



"""Module provides functions to initialize and manage the DuckDB."""

from contextlib import contextmanager
from typing import Generator

from sqlmodel import Session, SQLModel, create_engine

import lightly_purple.server.models  # noqa: F401, required for SQLModel to work properly


class DatabaseManager:
    """Manages database connections and ensures proper resource handling."""

    _instance = None

    def __new__(cls, db_file: str = "purple.db"):
        """Create a new instance of the DatabaseManager."""
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            # File-based DuckDB
            cls._instance.engine = create_engine(f"duckdb:///{db_file}")
            # Initialize tables
            SQLModel.metadata.create_all(cls._instance.engine)
        return cls._instance

    @contextmanager
    def session(self) -> Generator[Session, None, None]:
        """Create a new database session."""
        session = Session(self.engine)
        try:
            yield session
            session.commit()
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()


# Global instance
db_manager = DatabaseManager()


# For FastAPI dependency injection
def get_session():
    """Yield a new session for database operations."""
    with db_manager.session() as session:
        yield session



"""Example of how to use the DatasetLoader class."""

import os

from lightly_purple import DatasetLoader

# Create a DatasetLoader instance
loader = DatasetLoader()

# Define the path to the dataset (folder containing data.yaml)
dataset_path = os.getenv("DATASET_PATH", "path/to/your/dataset")

# Load YOLO dataset using data.yaml path
yolo_loader, dataset_id = loader.from_yolo(
    f"{dataset_path}/dataset.yaml",
    input_split=os.getenv("PURPLE_DATASET_SPLIT"),
)

loader.launch()



"""This module contains models with shared type dependencies."""

from datetime import datetime, timezone
from typing import List, Optional
from uuid import UUID, uuid4

from sqlalchemy import UniqueConstraint
from sqlmodel import Field, Relationship, SQLModel

from lightly_purple.server.models.annotation import AnnotationBase
from lightly_purple.server.models.annotation_label import AnnotationLabelBase
from lightly_purple.server.models.dataset import DatasetBase
from lightly_purple.server.models.sample import SampleBase
from lightly_purple.server.models.tag import TagBase


class AnnotationTagLink(SQLModel, table=True):
    """AnnotationTagLink links Annotation and Tag Many-to-Many."""

    annotation_id: Optional[UUID] = Field(
        default=None, foreign_key="annotations.annotation_id", primary_key=True
    )
    tag_id: Optional[UUID] = Field(
        default=None, foreign_key="tags.tag_id", primary_key=True
    )


class Annotation(AnnotationBase, table=True):
    """This class defines the Annotation model."""

    __tablename__ = "annotations"
    annotation_id: Optional[UUID] = Field(
        default_factory=uuid4, primary_key=True
    )
    annotation_label: "AnnotationLabel" = Relationship(
        back_populates="annotations",
        sa_relationship_kwargs={"lazy": "select"},
    )
    sample: Optional["Sample"] = Relationship(
        back_populates="annotations",
        sa_relationship_kwargs={"lazy": "select"},
    )

    """The tag ids associated with the sample."""
    tags: List["Tag"] = Relationship(
        back_populates="annotations", link_model=AnnotationTagLink
    )


class AnnotationLabel(AnnotationLabelBase, table=True):
    """This class defines the AnnotationLabel model."""

    __tablename__ = "annotation_labels"
    annotation_label_id: Optional[UUID] = Field(
        default_factory=uuid4, primary_key=True
    )
    annotations: List["Annotation"] = Relationship(
        back_populates="annotation_label",
    )


class SampleTagLink(SQLModel, table=True):
    """SampleTagLink links Sample and Tag Many-to-Many."""

    sample_id: Optional[UUID] = Field(
        default=None, foreign_key="samples.sample_id", primary_key=True
    )
    tag_id: Optional[UUID] = Field(
        default=None, foreign_key="tags.tag_id", primary_key=True
    )


class Sample(SampleBase, table=True):
    """This class defines the Sample model."""

    __tablename__ = "samples"
    sample_id: Optional[UUID] = Field(default_factory=uuid4, primary_key=True)
    created_at: datetime = Field(
        default_factory=lambda: datetime.now(timezone.utc)
    )
    updated_at: datetime = Field(
        default_factory=lambda: datetime.now(timezone.utc),
    )
    annotations: List["Annotation"] = Relationship(
        back_populates="sample",
    )

    """The tag ids associated with the sample."""
    tags: List["Tag"] = Relationship(
        back_populates="samples", link_model=SampleTagLink
    )


class Tag(TagBase, table=True):
    """This class defines the Tag model."""

    __tablename__ = "tags"
    # ensure there can only be one tag named "purple" per dataset
    __table_args__ = (
        UniqueConstraint("dataset_id", "name", name="unique_name_constraint"),
    )
    tag_id: Optional[UUID] = Field(default_factory=uuid4, primary_key=True)
    dataset_id: UUID
    created_at: datetime = Field(
        default_factory=lambda: datetime.now(timezone.utc)
    )
    updated_at: datetime = Field(
        default_factory=lambda: datetime.now(timezone.utc),
    )

    """The sample ids associated with the tag."""
    samples: List["Sample"] = Relationship(
        back_populates="tags", link_model=SampleTagLink
    )
    """The annotation ids associated with the tag."""
    annotations: List["Annotation"] = Relationship(
        back_populates="tags", link_model=AnnotationTagLink
    )


class Dataset(DatasetBase, table=True):
    """This class defines the Dataset model."""

    __tablename__ = "datasets"
    dataset_id: Optional[UUID] = Field(default_factory=uuid4, primary_key=True)
    created_at: datetime = Field(
        default_factory=lambda: datetime.now(timezone.utc)
    )
    updated_at: datetime = Field(
        default_factory=lambda: datetime.now(timezone.utc),
    )


class AnnotationView(SQLModel):
    """Annotation view model."""

    x: float
    y: float
    width: float
    height: float
    annotation_label: "AnnotationLabelView" = None


class AnnotationLabelView(AnnotationLabelBase):
    """AnnotationLabel class when retrieving."""

    annotation_label_id: UUID


class SampleView(SQLModel):
    """Sample class when retrieving."""

    """The name of the image file."""
    file_name: str
    file_path_abs: str
    sample_id: UUID
    annotations: List["AnnotationView"] = []  # noqa: RUF012



"""This module contains the API routes for managing tags."""

from typing import List, Optional
from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException, Path, Query
from pydantic import BaseModel
from sqlalchemy.exc import IntegrityError
from sqlmodel import Field, Session
from typing_extensions import Annotated

from lightly_purple.server.db import get_session
from lightly_purple.server.models import Tag
from lightly_purple.server.models.tag import (
    TagInput,
    TagInputBody,
    TagUpdate,
    TagUpdateBody,
    TagView,
)
from lightly_purple.server.resolvers.tag import TagResolver
from lightly_purple.server.routes.api.status import (
    HTTP_STATUS_CONFLICT,
    HTTP_STATUS_CREATED,
    HTTP_STATUS_NOT_FOUND,
)
from lightly_purple.server.routes.api.validators import Paginated

tag_router = APIRouter()
SessionDep = Annotated[Session, Depends(get_session)]


def get_tag_resolver(session: SessionDep) -> TagResolver:
    """Create an instance of the TagResolver."""
    return TagResolver(session)


@tag_router.post(
    "/datasets/{dataset_id}/tags",
    response_model=TagView,
    status_code=HTTP_STATUS_CREATED,
)
def create_tag(
    handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    dataset_id: Annotated[UUID, Path(title="Dataset Id")],
    body: TagInputBody,
) -> Tag:
    """Create a new tag in the database."""
    try:
        return handler.create(
            TagInput(
                **body.model_dump(exclude_unset=True), dataset_id=dataset_id
            )
        )
    except IntegrityError as e:
        raise HTTPException(
            status_code=HTTP_STATUS_CONFLICT,
            detail=f"""
                Tag with name {body.name} already exists
                in the dataset {dataset_id}.
            """,
        ) from e


@tag_router.get("/datasets/{dataset_id}/tags")
def read_tags(
    handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    dataset_id: Annotated[UUID, Path(title="Dataset Id")],
    paginated: Annotated[Paginated, Query()],
) -> List[TagView]:
    """Retrieve a list of tags from the database."""
    print("Get all tags, wut?")
    return handler.get_all_by_dataset_id(
        **paginated.model_dump(),
        dataset_id=dataset_id,
    )


@tag_router.get("/datasets/{dataset_id}/tags/{tag_id}")
def read_tag(
    handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    dataset_id: Annotated[UUID, Path(title="Dataset Id")],
    tag_id: Annotated[UUID, Path(title="Tag Id")],
) -> Tag:
    """Retrieve a single tag from the database."""
    tag = handler.get_by_id(tag_id)
    if not tag:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND,
            detail=f"""
            Tag with id {tag_id} for dataset {dataset_id} not found.
            """,
        )
    return tag


@tag_router.put("/datasets/{dataset_id}/tags/{tag_id}")
def update_tag(
    handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    dataset_id: Annotated[UUID, Path(title="Dataset Id")],
    tag_id: Annotated[UUID, Path(title="Tag Id")],
    body: TagUpdateBody,
) -> Tag:
    """Update an existing tag in the database."""
    try:
        tag = handler.update(
            tag_id,
            TagUpdate(
                **body.model_dump(exclude_unset=True), dataset_id=dataset_id
            ),
        )
        if not tag:
            raise HTTPException(
                status_code=HTTP_STATUS_NOT_FOUND,
                detail=f"Tag with id {tag_id} not found.",
            )
    except IntegrityError as e:
        raise HTTPException(
            status_code=HTTP_STATUS_CONFLICT,
            detail=f"""
                Cannot update tag. Tag with name {body.name}
                already exists in the dataset {dataset_id}.
            """,
        ) from e
    return tag


@tag_router.delete("/datasets/{dataset_id}/tags/{tag_id}")
def delete_tag(
    handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    tag_id: Annotated[UUID, Path(title="Tag Id")],
):
    """Delete a tag from the database."""
    if not handler.delete(tag_id):
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND, detail="tag not found"
        )
    return {"status": "deleted"}


class SampleIdsBody(BaseModel):
    """body parameters for adding or removing thing_ids."""

    sample_ids: Optional[List[UUID]] = Field(
        None, description="sample ids to add/remove"
    )


@tag_router.post(
    "/datasets/{dataset_id}/tags/{tag_id}/add/samples",
    status_code=HTTP_STATUS_CREATED,
)
def add_sample_ids_to_tag_id(
    tag_handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    tag_id: UUID,
    body: SampleIdsBody,
) -> bool:
    """Add sample_ids to a tag_id."""
    tag = tag_handler.get_by_id(tag_id)
    if not tag:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND,
            detail=f"Tag {tag_id} not found, can't add sample_ids.",
        )

    tag_handler.add_sample_ids_to_tag_id(tag_id, **body.model_dump())
    return True


@tag_router.delete(
    "/datasets/{dataset_id}/tags/{tag_id}/remove/samples",
)
def remove_thing_ids_to_tag_id(
    tag_handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    tag_id: UUID,
    body: SampleIdsBody,
) -> bool:
    """Add thing_ids to a tag_id."""
    tag = tag_handler.get_by_id(tag_id)
    if not tag:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND,
            detail=f"Tag {tag_id} not found, can't remove samples.",
        )

    tag_handler.remove_sample_ids_from_tag_id(tag_id, **body.model_dump())
    return True


class AnnotationIdsBody(BaseModel):
    """body parameters for adding or removing annotation_ids."""

    annotation_ids: Optional[List[UUID]] = Field(
        None, description="annotation ids to add/remove"
    )


@tag_router.post(
    "/datasets/{dataset_id}/tags/{tag_id}/add/annotations",
    status_code=HTTP_STATUS_CREATED,
)
def add_annotation_ids_to_tag_id(
    tag_handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    tag_id: UUID,
    body: AnnotationIdsBody,
) -> bool:
    """Add thing_ids to a tag_id."""
    tag = tag_handler.get_by_id(tag_id)
    if not tag:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND,
            detail=f"Tag {tag_id} not found, can't add annotations.",
        )

    tag_handler.add_annotation_ids_to_tag_id(tag_id, **body.model_dump())
    return True


@tag_router.delete(
    "/datasets/{dataset_id}/tags/{tag_id}/remove/annotations",
)
def remove_annotation_ids_to_tag_id(
    tag_handler: Annotated[TagResolver, Depends(get_tag_resolver)],
    tag_id: UUID,
    body: AnnotationIdsBody,
) -> bool:
    """Add thing_ids to a tag_id."""
    tag = tag_handler.get_by_id(tag_id)
    if not tag:
        raise HTTPException(
            status_code=HTTP_STATUS_NOT_FOUND,
            detail=f"Tag {tag_id} not found, can't remove annotations.",
        )

    tag_handler.remove_annotation_ids_from_tag_id(tag_id, **body.model_dump())
    return True



"""Dataset functionality module."""

import webbrowser
from pathlib import Path
from typing import Optional, Tuple
from uuid import UUID

from labelformat.model.bounding_box import BoundingBoxFormat
from tqdm import tqdm

from lightly_purple.dataset.env import APP_URL, PURPLE_HOST, PURPLE_PORT
from lightly_purple.server.db import db_manager
from lightly_purple.server.models import Dataset
from lightly_purple.server.models.annotation import AnnotationInput
from lightly_purple.server.models.annotation_label import AnnotationLabelInput
from lightly_purple.server.models.dataset import DatasetInput
from lightly_purple.server.models.sample import SampleInput
from lightly_purple.server.models.tag import TagInput
from lightly_purple.server.resolvers.annotation import AnnotationResolver
from lightly_purple.server.resolvers.annotation_label import AnnotationLabelResolver
from lightly_purple.server.resolvers.dataset import DatasetResolver
from lightly_purple.server.resolvers.sample import SampleResolver
from lightly_purple.server.resolvers.tag import TagResolver
from lightly_purple.server.server import Server

from .yolo_loader import YOLODatasetLoader

# Constants
ANNOTATION_BATCH_SIZE = 64  # Number of annotations to process in a single batch


class DatasetLoader:
    """Class responsible for loading datasets from various sources."""

    def __init__(self):
        """Initialize the dataset loader."""
        self._yolo_loader: Optional[YOLODatasetLoader] = None
        self._dataset: Optional[Dataset] = None
        with db_manager.session() as session:
            self.dataset_resolver = DatasetResolver(session)
            self.tag_resolver = TagResolver(session)
            self.sample_resolver = SampleResolver(session)
            self.annotation_resolver = AnnotationResolver(session)
            self.annotation_label_resolver = AnnotationLabelResolver(session)

    def from_yolo(
        self, data_yaml_path: str, input_split: str = "train"
    ) -> Tuple[YOLODatasetLoader, UUID]:
        """Load a dataset in YOLO format and store in database."""
        if not self._yolo_loader:
            self._yolo_loader = YOLODatasetLoader(data_yaml_path, input_split)

        # Load the dataset
        label_input = self._yolo_loader.load()

        with db_manager.session() as session:  # noqa: F841
            # Create dataset record
            dataset = DatasetInput(
                name=Path(data_yaml_path).parent.name,
                directory=str(Path(data_yaml_path).parent.absolute()),
            )

            self._dataset = self.dataset_resolver.create(dataset)

            # TODO(Kondrat 01/25): We need to expose images_dir from label_input
            img_dir = self._yolo_loader._label_input._images_dir()  # noqa: SLF001

            # Store labels first
            label_map = {}
            for category in tqdm(
                label_input.get_categories(), desc="Processing categories"
            ):
                label = AnnotationLabelInput(
                    annotation_label_name=category.name
                )
                stored_label = self.annotation_label_resolver.create(label)
                label_map[category.id] = stored_label.annotation_label_id

            annotations_to_create = []

            # temporary hack; create dummy tags until we can create tags
            tag_even = self.tag_resolver.create(
                TagInput(dataset_id=self._dataset.dataset_id, name="even")
            )
            tag_mod5 = self.tag_resolver.create(
                TagInput(dataset_id=self._dataset.dataset_id, name="mod5")
            )

            # Process images and annotations
            for i, image_data in enumerate(tqdm(
                label_input.get_labels(), desc="Processing images"
            )):
                # Create sample record
                sample = SampleInput(
                    file_name=str(image_data.image.filename),
                    file_path_abs=str(img_dir / image_data.image.filename),
                    width=image_data.image.width,
                    height=image_data.image.height,
                    dataset_id=self._dataset.dataset_id,
                )
                stored_sample = self.sample_resolver.create(sample)

                # temporary hack; create dummy tags until we can create tags
                if (i % 2) == 0:
                    self.tag_resolver.add_tag_to_sample(
                        tag_even.tag_id, stored_sample
                    )
                if (i % 5) == 0:
                    self.tag_resolver.add_tag_to_sample(
                        tag_mod5.tag_id, stored_sample
                    )

                # Create annotations
                for obj in image_data.objects:
                    box = obj.box.to_format(BoundingBoxFormat.XYWH)
                    x, y, width, height = box

                    annotations_to_create.append(
                        AnnotationInput(
                            dataset_id=self._dataset.dataset_id,
                            sample_id=stored_sample.sample_id,
                            annotation_label_id=label_map[obj.category.id],
                            x=x,
                            y=y,
                            width=width,
                            height=height,
                        )
                    )

                if len(annotations_to_create) >= ANNOTATION_BATCH_SIZE:
                    self.annotation_resolver.create_many(annotations_to_create)
                    annotations_to_create = []

            # Insert any remaining annotations
            if annotations_to_create:
                self.annotation_resolver.create_many(annotations_to_create)
                annotations_to_create = []

        # TODO: we should not return internal state but use getters
        return self._yolo_loader, self._dataset.dataset_id

    def launch(self):
        """Launch the web interface for the loaded dataset."""
        server = Server(host=PURPLE_HOST, port=int(PURPLE_PORT))

        print(f"Opening URL: {APP_URL}")

        # We need to open browser before starting the server
        webbrowser.open_new(APP_URL)

        server.start()
